#ifndef lint
static char *RCSid() { return RCSid("$Id: amavis.c,v 1.1.2.27 2002/05/31 11:05:19 lhecking Exp $"); }
#endif

/*
 * client for amavisd
 *
 * Author: Geoff Winkless <gwinkless@users.sourceforge.net>
 * Additional work and patches by:
 *   Gregory Ade
 *   Thomas Biege
 *   Pierre-Yves Bonnetain
 *   Ricardo M. Ferreira
 *   Lars Hecking
 *   Rainer Link
 *   Julio Sanchez
 *
 */

/*
 * Add some copyright notice here ...
 *
 * Usage: amavis sender recipient [recipient ...] [-- lda [lda-args]]
 * Usage with qmail: no cmd line args
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define BUFFLEN 8192
/* Must be the same as the buffer length for recv() in amavisd */
#define SOCKBUFLEN 8192

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#define D_TEMPLATE "/amavis-XXXXXXXX"
#define F_TEMPLATE "/email.txt"

#define QMAILQUEUE "qmail-queue"

#define NEXT_ENVELOPE(a,b) (isqmail? next_qmail_envelope((a),(b)) : \
	(i < argc? argv[i++] : NULL))

#define MSG_INSTREAM stdin
#define ENV_INSTREAM stdout

#define DBG_NONE  0
#define DBG_INFO  1
#define DBG_WARN  2
#define DBG_FATAL 4

#define DBG_ALL     (DBG_FATAL | DBG_WARN | DBG_INFO)

static struct utsname myuts;

/* Don't debug by default */
static const int debuglevel = DBG_NONE;
/* static const int debuglevel = DBG_ALL; */
static const char *mydebugfile = RUNTIME_DIR "/amavis.client";

/* temp dir where mail message is stored */
static char *dirname;
/* temp file where mail message is stored */
static char *atmpfile;
/* temp buffer for mail message - BUFFLEN */
static char *buff;

static void mydebug(const int, const char *, ...);
static char *mymktempdir(char *);
static char *next_qmail_envelope(char *, char *);
static void amavis_cleanup(void);


#ifndef HAVE_STRLCPY
static size_t
strlcpy(char *dst, const char *src, size_t size)
{
    size_t src_l = strlen(src);
    if(src_l < size) {
	memcpy(dst, src, src_l + 1);
    } else if(size) {
	memcpy(dst, src, size - 1);
	dst[size - 1] = '\0';
    }
    return src_l;
}
#endif


static void
mydebug(const int level, const char *fmt, ...)
{
    FILE *f = NULL;
    time_t tmpt;
    char *timestamp;
    int rv;
    va_list ap;

    if (!(level & debuglevel))
	return;

    /* Set up debug file */
    if (mydebugfile && (f = fopen(mydebugfile, "a")) == NULL) {
	fprintf(stderr, "error opening '%s': %s\n", mydebugfile, strerror(errno));
	return;
    }

    tmpt = time(NULL);
    timestamp = ctime(&tmpt);
    /* A 26 character string according ctime(3c)
     * we cut off the trailing \n\0 */
    timestamp[24] = 0;
    rv = fprintf(f, "%s %s amavis(client)[%ld]: ", timestamp,
		 (myuts.nodename ? myuts.nodename : "localhost"), (long) getpid());
    if (rv < 0)
	perror("error writing (fprintf) to debug file");

    va_start(ap, fmt);
    rv = vfprintf(f, fmt, ap);
    va_end(ap);
    if (rv < 0)
	perror("error writing (vfprintf) to debug file");

    rv = fputc('\n', f);
    if (rv < 0)
	perror("error writing (fputc) to debug file");
    rv = fclose(f);
    if (rv < 0)
	perror("error closing debug file f");
}


static char *
mymktempdir(char *s)
{
    char *stt;
    int count = 0;

#ifdef HAVE_MKDTEMP
    return mkdtemp(s);
#else
    /* magic number alert */
    while (count++ < 20) {
# ifdef HAVE_MKTEMP
	stt = mktemp(s);
# else
	/* relies on template format */
	stt = strchr(s, '-') + 1;
	if (stt) {
	    /* more magic number alert */
	    snprintf(stt, strlen(s) - 1 - (stt - s), "%08d", lrand48() / 215);
	    stt = s;
	} else {
	    /* invalid template */
	    return NULL;
	}
# endif
	if (stt) {
	    if (!mkdir(s, S_IRWXU)) {
		return s;
	    } else {
		continue;
	    }
	}
    }
    return NULL;
#endif /* HAVE_MKDTEMP */
}

static char *
next_qmail_envelope(char *first, char *last)
{
    static char *p = NULL, *q;

    if (!p) {
	p = first;
	q = last;
	return p;
    }

    /* get args from first */
    if (p < q) {
	if (*p == 'T' || *p == 'F') {
	    return p+1;
	} else {
	    mydebug(DBG_FATAL, "qmail envelope format error (F/T)");
	    exit(EX_TEMPFAIL);
	}
	p += strlen(p) + 1;
    }

    return NULL;
}

static void
amavis_cleanup(void)
{
    if (dirname)
	free(dirname);

    if (atmpfile)
	free(atmpfile);

    if (buff)
	free(buff);
}

/* Simple "protocol" */
const char _LDA = '\2';
const char _EOT = '\3';

/* take input from stdin as an email, with argv as sender and recipients
 * then pass it all to amavisd
 * TODO: sender/recipient parsing for qmail; see qmail-queue(8) for details
 */
int
main(int argc, char **argv)
{
    char xstat[8] = { 0 };
    struct sockaddr_un saddr;
    FILE *fout = NULL;
    int fd = 0;
    const char *envelope;
    size_t rw = 0, msgsize = 0;
    int r, sock, i;		/* socket func return val, socket descriptor, index var */
    int isqmail = 0;
    char retval;
    struct stat StatBuf;
    char *p, *q;		/* for qmail */
#if !(HAVE_MKDTEMP || HAVE_MKTEMP)
    int mypid = getpid();

    srand48(time(NULL) ^ (mypid + (mypid << 15)));
#endif

    atexit(amavis_cleanup);

    if (strstr(argv[0], QMAILQUEUE) != NULL)
	isqmail++;

    /* */
    uname(&myuts);

    /* Process args first */
    if (!isqmail && argc < 3) {
	mydebug(DBG_FATAL, "Insufficient number of arguments: %s", strerror(errno));
	exit(EX_TEMPFAIL);
    }

    umask(0077);

    /* */
    dirname = malloc(strlen(RUNTIME_DIR) + strlen(D_TEMPLATE) + 1);
    if (dirname == NULL) {
	mydebug(DBG_FATAL, "Failed to allocate memory for temp dir name: %s", strerror(errno));
	exit(EX_TEMPFAIL);
    }

    strcpy(dirname, RUNTIME_DIR);
    strcat(dirname, D_TEMPLATE);
    if (mymktempdir(dirname) == NULL) {
	mydebug(DBG_FATAL, "Failed to create temp dir: %s", strerror(errno));
	exit(EX_TEMPFAIL);
    }

    if (lstat(dirname, &StatBuf) < 0) {
	mydebug(DBG_FATAL, "%s: Error while trying lstat(%s): %s",
		argv[0], dirname, strerror(errno));
	exit(EX_TEMPFAIL);
    }

    /* may be too restrictive for you, but's good to avoid problems */
    if (!S_ISDIR(StatBuf.st_mode) || StatBuf.st_uid != geteuid() ||
	StatBuf.st_gid != getegid() || !(StatBuf.st_mode & (S_IWUSR | S_IRUSR))) {
	mydebug(DBG_FATAL,
		"%s: Security Warning: %s must be a Directory and owned by "
		"User %d and Group %d and just read-/write-able by the User "
		" and noone else. Exit.", argv[0], dirname, geteuid(), getegid());
	exit(EX_TEMPFAIL);
    }
    /* there is still a race condition here if RUNTIME_DIR is writeable by the attacker :-\ */

    atmpfile = malloc(strlen(dirname) + strlen(F_TEMPLATE) + 1);
    if (atmpfile == NULL) {
	mydebug(DBG_FATAL, "Failed to allocate memory for temp file name: %s", strerror(errno));
	exit(EX_TEMPFAIL);
    }

    sprintf(atmpfile, "%s/email.txt", dirname);

    buff = malloc(BUFFLEN);
    if (buff == NULL) {
	mydebug(DBG_FATAL, "Failed to allocate memory for read buffer: %s", strerror(errno));
	exit(EX_TEMPFAIL);
    }

    if ((fd = open(atmpfile, O_CREAT | O_EXCL | O_WRONLY, S_IRUSR)) < 0 || (fout = fdopen(fd, "w")) == NULL)
	mydebug(DBG_FATAL, "failed to open a_tmp_file: %s", strerror(errno));

    while (!feof(MSG_INSTREAM)) {
	rw = fread(buff, sizeof(char), BUFFLEN, MSG_INSTREAM);
	fwrite(buff, sizeof(char), rw, fout);
	msgsize += rw;
    }
    fclose(fout);
    mydebug(DBG_INFO, "size=%d", msgsize);

    if (isqmail) {
#define GROWFACTOR 2
	size_t obsize = 0, nbsize;

	/* reuse buff for qmail envelope */
	memset(buff, 0, BUFFLEN);
	nbsize = BUFFLEN;
	p = buff;

	/* read envelope from stdout (!) */
	while(feof(ENV_INSTREAM) == 0) {
	    rw = fread(p, sizeof(char), nbsize-obsize, ENV_INSTREAM);
	    if (ferror(ENV_INSTREAM) != 0) {
		perror("read failed");
		exit(EX_TEMPFAIL);
	    }
	    if (rw == (nbsize-obsize)) {
		obsize = nbsize;
		nbsize *= GROWFACTOR;
		q = realloc(buff, nbsize);
		if (q == NULL) {
		    perror("realloc failed");
		    exit(EX_TEMPFAIL);
		}
		buff = q;
		p = buff + obsize;
		memset(p, 0, nbsize-obsize);
	    } else
		p += rw;
	}

	/* point q to the trailing extra 0 */
	q = p - 1;
	if (*q != '\0')
	    perror("Format error");

	/* init next_qmail_envelope()'s internal static vars */
	if (next_qmail_envelope(buff,q) == NULL) {
	    perror("Internal error in next_qmail_envelope");
	    exit(EX_TEMPFAIL);
	}
    }

    r = (sock = socket(PF_UNIX, SOCK_STREAM, 0));
    if (r < 0)
	mydebug(DBG_FATAL, "failed to allocate socket: %s", strerror(errno));
    saddr.sun_family = AF_UNIX;
    strlcpy(saddr.sun_path, AMAVISD_SOCKET, sizeof(saddr.sun_path));
    if (r >= 0) {
	mydebug(DBG_INFO, "connect()");
	r = connect(sock, (struct sockaddr *) &saddr, sizeof(saddr));
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to connect(): %s", strerror(errno));
    }
    if (r >= 0) {
	mydebug(DBG_INFO, "senddir() %s", dirname);
	r = send(sock, dirname, strlen(dirname), 0);
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to send() directory: %s", strerror(errno));
    }
    if (r >= 0) {
	r = recv(sock, &retval, 1, 0);
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to recv() directory confirmation: %s", strerror(errno));
    }

    /* index into argv */
    i = 1;

    if (r >= 0) {
	/* send envelope from */
	int sender_l;
	envelope = NEXT_ENVELOPE(buff,q);
	envelope = (strlen(envelope) > 0) ? envelope : "<>";
	mydebug(DBG_INFO, "sendfrom() %s", envelope);
	sender_l = strlen(envelope);
	if (sender_l > SOCKBUFLEN) {
/* FIXME: envelope is in buff, which may be longer than SOCKBUFLEN.
   However, amavisd still has a static SOCKBUFLEN limit */
	    mydebug(DBG_WARN, "Sender too long (%d), truncated to %d characters", sender_l, SOCKBUFLEN);
	    sender_l = SOCKBUFLEN;
	}
	r = send(sock, envelope, sender_l, 0);
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to send() Sender: %s", strerror(errno));
	else if (r < sender_l)
	    mydebug(DBG_WARN, "failed to send() complete Sender, truncated to %d characters ", r);
    }
    if (r >= 0) {
	r = recv(sock, &retval, 1, 0);
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to recv() ok for Sender info: %s", strerror(errno));
    }
    if (r >= 0) {
	/* send recipients and lda/ldaargs if present */
	while (1) {
	    int arg_l;
	    envelope = NEXT_ENVELOPE(buff,q);
	    if (envelope == NULL)
		break;

	    mydebug(DBG_INFO, "next: (%s), %d", envelope, i);

	    arg_l = strlen(envelope);
	    if (arg_l > SOCKBUFLEN) {
		mydebug(DBG_WARN, "Recipient too long (%d), truncated to %d characters", arg_l, SOCKBUFLEN);
		arg_l = SOCKBUFLEN;
	    }
	    if (strcmp(envelope, "--") == 0) {
		/* cheat ... */
		arg_l = 1;
		mydebug(DBG_INFO, "sendlda() %s", envelope);
		r = send(sock, &_LDA, 1, 0);
	    } else {
		mydebug(DBG_INFO, "sendto() %s", envelope);
		/* mydebug(DBG_INFO, "sendlda/arg() %s", envelope); */
		r = send(sock, envelope, arg_l, 0);
	    }
/* FIXME: clean up debug messages */
	    if (r < 0) {
		mydebug(DBG_FATAL, "failed to send() Recipient: %s", strerror(errno));
		/* mydebug(DBG_FATAL, "failed to send/lda() recip info: %s", strerror(errno)); */
	    } else {
		if (r < arg_l)
		    mydebug(DBG_WARN, "failed to send() complete Recipient, truncated to %d characters", r);
		r = recv(sock, &retval, 1, 0);
		if (r < 0) {
		    mydebug(DBG_FATAL, "failed to recv() ok for recip info: %s", strerror(errno));
		    /* mydebug(DBG_FATAL, "failed to recv() ok for recip/lda info: %s", strerror(errno)); */
		}
	    }
	}
    }
    if (r >= 0) {
	/* send end of transmission */
	mydebug(DBG_INFO, "sendEOT()");
	r = send(sock, &_EOT, 1, 0);
	if (r < 0)
	    mydebug(DBG_FATAL, "failed to send() EOT: %s", strerror(errno));
	else {
	    r = recv(sock, xstat, 6, 0);
	    mydebug(DBG_INFO, "received %s from daemon", xstat);
	    if (r < 0)
		mydebug(DBG_FATAL, "Failed to recv() final result: %s", strerror(errno));
	    else if (!r)
		mydebug(DBG_FATAL, "Failed to recv() final result: empty status string");
	    /* get back final result */
	}
    }
    close(sock);
    mydebug(DBG_INFO, "finished conversation");
    if (r < 0) {
	/* housekeeping */
	unlink(atmpfile);
	rmdir(dirname);
	mydebug(DBG_FATAL, "failing with EX_TEMPFAIL: %s", strerror(errno));
	exit(EX_TEMPFAIL);
	/* some point of the communication failed miserably - so give up */
    }

    /* Protect against empty return string */
    if (*xstat)
	retval = atoi(xstat);
    else
	retval = 1;

    mydebug(DBG_INFO, "retval is %d\n", retval);

    if (retval)
	exit(EX_TEMPFAIL);

    exit(retval);
}
